from django import dispatch
from django.db import transaction
from django.db.models import signals as model_signals

from asgiref.sync import async_to_sync
from channels.exceptions import ChannelFull
from channels.layers import get_channel_layer

from .models import Observer, Subscriber
from .protocol import *


def notify_observers(table, kind, primary_key=None):
    """Transmit ORM table change notifcation.

    :param table: Name of the table that has changed
    :param kind: Change type
    :param primary_key: Primary key of the affected instance
    """

    try:
        async_to_sync(get_channel_layer().send)(
            CHANNEL_WORKER_NOTIFY,
            {
                'type': TYPE_ORM_NOTIFY_TABLE,
                'table': table,
                'kind': kind,
                'primary_key': primary_key,
            }
        )
    except ChannelFull:
        pass


@dispatch.receiver(model_signals.post_save)
def model_post_save(sender, instance, created=False, **kwargs):
    """Signal emitted after any model is saved via Django ORM.

    :param sender: Model class that was saved
    :param instance: The actual instance that was saved
    :param created: True if a new row was created
    """

    def notify():
        table = sender._meta.db_table
        if created:
            notify_observers(table, ORM_NOTIFY_KIND_CREATE, instance.pk)
        else:
            notify_observers(table, ORM_NOTIFY_KIND_UPDATE, instance.pk)

    transaction.on_commit(notify)


@dispatch.receiver(model_signals.post_delete)
def model_post_delete(sender, instance, **kwargs):
    """Signal emitted after any model is deleted via Django ORM.

    :param sender: Model class that was deleted
    :param instance: The actual instance that was removed
    """

    def notify():
        table = sender._meta.db_table
        notify_observers(table, ORM_NOTIFY_KIND_DELETE, instance.pk)

    transaction.on_commit(notify)


@dispatch.receiver(model_signals.m2m_changed)
def model_m2m_changed(sender, instance, action, **kwargs):
    """
    Signal emitted after any M2M relation changes via Django ORM.

    :param sender: M2M intermediate model
    :param instance: The actual instance that was saved
    :param action: M2M action
    """

    def notify():
        table = sender._meta.db_table
        if action == 'post_add':
            notify_observers(table, ORM_NOTIFY_KIND_CREATE)
        elif action in ('post_remove', 'post_clear'):
            notify_observers(table, ORM_NOTIFY_KIND_DELETE)

    transaction.on_commit(notify)


@dispatch.receiver(model_signals.post_delete, sender=Subscriber)
def subscriber_removed(sender, instance, **kwargs):
    """Remove observer if all subscribers removed.

    Ensure that when all subscribers are removed from an observer, the
    observer itself is also removed.

    """
    Observer.objects.filter(subscribers=None).delete()



import collections
import pickle
import json
import logging
import sys
import time

import six

from django.core import exceptions as django_exceptions
from django.db import transaction, IntegrityError
from django.http import Http404
from django.utils import timezone

from rest_framework import request as api_request

from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer

from . import exceptions, models
from .connection import get_queryobserver_settings
from .interceptor import QueryInterceptor
from .protocol import *

# Logger.
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

# Observable method options attribute name prefix.
OBSERVABLE_OPTIONS_PREFIX = 'observable_'


class Options(object):
    """Query observer options."""

    # Valid change detection types.
    CHANGE_DETECTION_PUSH = 'push'
    CHANGE_DETECTION_POLL = 'poll'

    def __init__(self, viewset, viewset_method):
        self._viewset = viewset
        self._viewset_method = viewset_method

        # Determine the primary key.
        self.primary_key = self.get_option('primary_key')
        if self.primary_key is None:
            # Primary key attribute is not defined, attempt to autodiscover it from the queryset.
            try:
                self.primary_key = viewset.get_queryset().model._meta.pk.name
            except AssertionError:
                # No queryset is defined.
                raise exceptions.MissingPrimaryKey(
                    "Observable method does not define a primary key and the viewset "
                    "does not provide a queryset. Define a queryset or use the primary_key "
                    "decorator."
                )

        # Determine change detection type.
        self.change_detection = self.get_option('change_detection', Options.CHANGE_DETECTION_PUSH)
        self.poll_interval = self.get_option('poll_interval')

    def get_option(self, name, default=None):
        return getattr(self._viewset_method, '{}{}'.format(OBSERVABLE_OPTIONS_PREFIX, name), default)


class QueryObserver(object):
    """Query observer.

    A query observer observes a specific viewset for changes and propagates these
    changes to all interested subscribers.
    """

    def __init__(self, request):
        """Create new query observer.

        :param request: A `queryobserver.request.Request` instance
        """

        # Obtain a serializer by asking the viewset to provide one. We instantiate the
        # viewset with a fake request, so that the viewset methods work as expected.
        viewset = request.viewset_class()
        viewset.request = api_request.Request(request)
        viewset.request.method = request.method
        viewset.format_kwarg = None
        viewset.args = request.args
        viewset.kwargs = request.kwargs
        self._viewset = viewset
        self._request = request
        self._viewset_method = getattr(viewset, request.viewset_method)
        self._meta = Options(viewset, self._viewset_method)

    @property
    def id(self):
        """Unique observer identifier."""
        return self._request.observe_id

    def _get_logging_extra(self, stopped=False, duration=None, results=None):
        """Extra information for logger."""
        return {
            'stopped': stopped,
            'duration': duration,
            'results': results,
            'observer_id': self.id,
            'viewset': '{}.{}'.format(
                self._request.viewset_class.__module__,
                self._request.viewset_class.__name__
            ),
            'method': self._request.viewset_method,
            'path': self._request.path,
            'get': self._request.GET,
        }

    def _get_logging_id(self):
        """Get logging identifier."""
        return "{}.{}/{}".format(
            self._request.viewset_class.__module__,
            self._request.viewset_class.__name__,
            self._request.viewset_method,
        )

    def evaluate(self, return_full=True, return_emitted=False):
        """Evaluate the query observer.

        :param return_full: True if the full set of rows should be returned
        :param return_emitted: True if the emitted diffs should be returned
        """

        try:
            settings = get_queryobserver_settings()

            with transaction.atomic():
                # Obtain the observer state and lock it. This prevents an observer from being
                # processed in parallel from multiple different workers.
                try:
                    observer = models.Observer.objects.select_for_update().get(pk=self.id)
                except models.Observer.DoesNotExist:
                    # No observer yet, create one.
                    observer = models.Observer.objects.create(
                        id=self.id,
                        request=pickle.dumps(self._request),
                    )

                # Evaluate the observer.
                start = time.time()
                result = self._evaluate(observer, return_full, return_emitted)
                duration = time.time() - start

                # Log slow observers.
                if duration > settings['warnings']['max_processing_time']:
                    logger.warning(
                        "Slow observed viewset ({})".format(self._get_logging_id()),
                        extra=self._get_logging_extra(duration=duration)
                    )

                # Stop really slow observers.
                if duration > settings['errors']['max_processing_time']:
                    logger.error(
                        "Stopped extremely slow observed viewset ({})".format(self._get_logging_id()),
                        extra=self._get_logging_extra(stopped=True, duration=duration)
                    )
                    observer.delete()

            return result
        except:  # pylint: disable=bare-except
            logger.exception(
                "Error while evaluating observer ({})".format(self._get_logging_id()),
                extra=self._get_logging_extra()
            )
            return []

    def _evaluate(self, observer, return_full=True, return_emitted=False):
        """Evaluates the query observer.

        This method must be run in a transaction with the `observer` locked
        for update.

        :param observer: Observer state model instance
        :param return_full: True if the full set of rows should be returned
        :param return_emitted: True if the emitted diffs should be returned
        """
        # Evaluate the viewset, intercepting all queries to evaluate dependencies.
        tables = set()
        stop_observer = False
        with QueryInterceptor().intercept(tables):
            try:
                response = self._viewset_method(
                    self._viewset.request,
                    *self._request.args,
                    **self._request.kwargs
                )

                if response.status_code == 200:
                    results = response.data

                    if not isinstance(results, list):
                        if isinstance(results, dict):
                            if 'results' in results and isinstance(results['results'], list):
                                # Support paginated results.
                                results = results['results']
                            else:
                                results.setdefault(self._meta.primary_key, 1)
                                results = [collections.OrderedDict(results)]
                        else:
                            raise ValueError(
                                "Observable views must return a dictionary or a list of dictionaries!"
                            )
                else:
                    results = []
            except Http404:
                results = []
            except django_exceptions.ObjectDoesNotExist:
                # The evaluation may fail when certain dependent objects (like users) are removed
                # from the database. In this case, the observer is stopped.
                stop_observer = True

        # Check if observer should be stopped.
        if stop_observer:
            observer.delete()
            return []

        # Determine who should notify us based on the configured change detection mechanism.
        if self._meta.change_detection == Options.CHANGE_DETECTION_PUSH:
            # Register table dependencies for push observables.
            for table in tables:
                models.Dependency.objects.get_or_create(
                    observer=observer,
                    table=table,
                )
        elif self._meta.change_detection == Options.CHANGE_DETECTION_POLL:
            # Register poller.
            observer.poll_interval = self._meta.poll_interval

            async_to_sync(get_channel_layer().send)(
                CHANNEL_POLL_OBSERVER,
                {
                    'type': TYPE_POLL_OBSERVER,
                    'observer': self.id,
                    'interval': self._meta.poll_interval,
                },
            )
        else:
            raise NotImplementedError("Change detection mechanism '{}' not implemented.".format(
                self._meta.change_detection
            ))

        # Update last evaluation time.
        observer.last_evaluation = timezone.now()
        observer.save()

        # Log viewsets with too much output.
        max_result_length = get_queryobserver_settings()['warnings']['max_result_length']
        if len(results) > max_result_length:
            logger.warning(
                "Observed viewset returned too many results ({})".format(self._get_logging_id()),
                extra=self._get_logging_extra(results=len(results))
            )

        new_results = collections.OrderedDict()
        for order, item in enumerate(results):
            if not isinstance(item, dict):
                raise ValueError("Observable views must return a dictionary or a list of dictionaries!")

            item = {
                'order': order,
                'data': item,
            }

            try:
                new_results[str(item['data'][self._meta.primary_key])] = item
            except KeyError:
                raise KeyError(
                    "Observable view did not return primary key field '{}'!".format(self._meta.primary_key)
                )

        # Process difference between old results and new results.
        added = []
        changed = []
        removed = []

        new_ids = list(new_results.keys())
        removed_qs = observer.items.exclude(primary_key__in=new_results.keys())
        maybe_changed_qs = observer.items.filter(primary_key__in=new_results.keys())

        # Removed items.
        removed = list(removed_qs.values('order', 'data'))
        removed_qs.delete()

        # Changed items.
        for item_id, old_order, old_data in maybe_changed_qs.values_list('primary_key', 'order', 'data'):
            new_item = new_results[item_id]
            new_ids.remove(item_id)

            if new_item['data'] != old_data:
                changed.append(new_item)
                observer.items.filter(primary_key=item_id).update(data=new_item['data'], order=new_item['order'])
            elif new_item['order'] != old_order:
                # TODO: If only order has changed, don't transmit full data (needs frontend support).
                changed.append(new_item)
                observer.items.filter(primary_key=item_id).update(order=new_item['order'])

        # Added items.
        for item_id in new_ids:
            item = new_results[item_id]
            added.append(item)
            observer.items.create(
                primary_key=item_id,
                order=item['order'],
                data=item['data'],
            )

        # Check whether to emit results.
        if observer.status == models.Observer.STATUS_OBSERVING:
            message = {
                'type': TYPE_ITEM_UPDATE,
                'observer': self.id,
                'primary_key': self._meta.primary_key,
                'added': added,
                'changed': changed,
                'removed': removed,
            }

            for subscriber in observer.subscribers.all():
                async_to_sync(get_channel_layer().group_send)(
                    GROUP_SESSIONS.format(session_id=subscriber.session_id),
                    message,
                )

            if return_emitted:
                return (added, changed, removed)
        else:
            # Switch observer status to OBSERVING.
            observer.status = models.Observer.STATUS_OBSERVING
            observer.save()

        if return_full:
            return [item['data'] for item in new_results.values()]

    def __eq__(self, other):
        return self.id == other.id

    def __hash__(self):
        return hash(self.id)

    def __repr__(self):
        return '<QueryObserver: id={id} request={request}>'.format(
            id=self.id,
            request=repr(self._request)
        )


def add_subscriber(session_id, observer_id):
    """Add subscriber to the given observer.

    :param session_id: Subscriber's session identifier
    :param observer_id: Observer identifier
    """
    with transaction.atomic():
        try:
            subscriber = models.Subscriber.objects.get(session_id=session_id)
        except models.Subscriber.DoesNotExist:
            # Subscriber does not exist as it may have already disconnected.
            return

        observer = models.Observer.objects.get(pk=observer_id)
        observer.subscribers.add(subscriber)


def remove_subscriber(session_id, observer_id):
    """Remove subscriber from the given observer.

    :param session_id: Subscriber's session identifier
    :param observer_id: Observer identifier
    """
    with transaction.atomic():
        try:
            subscriber = models.Subscriber.objects.get(session_id=session_id)
        except models.Subscriber.DoesNotExist:
            # Subscriber does not exist as it may have already disconnected.
            return

        observer = models.Observer.objects.get(pk=observer_id)
        observer.subscribers.remove(subscriber)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-05-02 04:21
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='ExampleItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('enabled', models.BooleanField()),
                ('name', models.CharField(max_length=30)),
            ],
            options={
                'ordering': ['pk'],
                'permissions': (('view_exampleitem', 'Can view example item'),),
            },
        ),
        migrations.CreateModel(
            name='ExampleM2MItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.IntegerField(default=1)),
                ('items', models.ManyToManyField(related_name='items', to='rest_framework_reactive_tests.ExampleItem')),
            ],
            options={
                'ordering': ['pk'],
            },
        ),
        migrations.CreateModel(
            name='ExampleSubItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('enabled', models.BooleanField()),
                ('parent', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='rest_framework_reactive_tests.ExampleItem')),
            ],
            options={
                'ordering': ['pk'],
                'permissions': (('view_examplesubitem', 'Can view example sub item'),),
            },
        ),
    ]



import json
import pickle

from django import test
from django.contrib.auth import models as auth_models

from guardian import shortcuts
from rest_framework import test as api_test, request as api_request

from . import models, views
from rest_framework_reactive import models as observer_models
from rest_framework_reactive import request as observer_request
from rest_framework_reactive.observer import QueryObserver

# Create test request factory.
factory = api_test.APIRequestFactory()


class QueryObserversTestCase(test.TestCase):
    def request(self, viewset_class, **kwargs):
        request = observer_request.Request(
            viewset_class,
            'list',
            api_request.Request(factory.get('/', kwargs))
        )

        # Simulate serialization.
        return pickle.loads(pickle.dumps(request))

    def test_paginated_viewset(self):
        observer = QueryObserver(self.request(views.PaginatedViewSet, offset=0, limit=10))
        items = observer.evaluate()

        self.assertEquals(len(items), 0)

        items = []
        for index in range(20):
            items.append(models.ExampleItem.objects.create(name='Example', enabled=True))

        # Evaluate the observer again (in reality this would be done automatically, triggered by signals
        # from Django ORM).
        added, changed, removed = observer.evaluate(return_emitted=True)

        self.assertEquals(len(added), 10)
        expected_serialized_item = {'id': items[0].pk, 'name': items[0].name, 'enabled': items[0].enabled}
        self.assertEquals(added[0]['data'], expected_serialized_item)
        self.assertEquals(len(changed), 0)
        self.assertEquals(len(removed), 0)

    def test_item_serialization(self):
        item = models.ExampleItem.objects.create(name='Example', enabled=True)
        shortcuts.assign_perm('rest_framework_reactive.view_exampleitem', auth_models.AnonymousUser(), item)

        observer = QueryObserver(self.request(views.ExampleItemViewSet))
        items = observer.evaluate()

        # Ensure items can be serialized into JSON.
        json.dumps(items)

    def test_observe_viewset(self):
        # Create a request and an observer for it.
        observer = QueryObserver(self.request(views.ExampleItemViewSet))
        items = observer.evaluate()

        self.assertEquals(observer.id, 'fdd1312a8082540528908c32f4a94cac55365ef7acadc8e5ae8d4795cd7b5fa6')
        self.assertEquals(len(items), 0)

        # Add an item into the database.
        item = models.ExampleItem()
        item.name = 'Example'
        item.enabled = True
        item.save()

        shortcuts.assign_perm('rest_framework_reactive.view_exampleitem', auth_models.AnonymousUser(), item)

        # Evaluate the observer again (in reality this would be done automatically, triggered by signals
        # from Django ORM).
        added, changed, removed = observer.evaluate(return_emitted=True)

        self.assertEquals(len(added), 1)
        expected_serialized_item = {'id': item.pk, 'name': item.name, 'enabled': item.enabled}
        self.assertEquals(added[0]['data'], expected_serialized_item)
        self.assertEquals(len(changed), 0)
        self.assertEquals(len(removed), 0)

        # Change the existing item.
        item.enabled = False
        expected_serialized_item['enabled'] = False
        item.save()

        added, changed, removed = observer.evaluate(return_emitted=True)
        self.assertEquals(len(added), 0)
        self.assertEquals(len(changed), 1)
        self.assertEquals(changed[0]['data'], expected_serialized_item)
        self.assertEquals(len(removed), 0)

        # Remove the first item.
        item.delete()

        # Add another two items.
        item2 = models.ExampleItem()
        item2.name = 'Example 2'
        item2.enabled = True
        item2.save()

        item3 = models.ExampleItem()
        item3.name = 'Example 3'
        item3.enabled = True
        item3.save()

        shortcuts.assign_perm('rest_framework_reactive.view_exampleitem', auth_models.AnonymousUser(), item2)
        shortcuts.assign_perm('rest_framework_reactive.view_exampleitem', auth_models.AnonymousUser(), item3)

        added, changed, removed = observer.evaluate(return_emitted=True)
        self.assertEquals(len(added), 2)
        self.assertEquals(added[0]['data'], {'id': item2.pk, 'name': item2.name, 'enabled': item2.enabled})
        self.assertEquals(added[1]['data'], {'id': item3.pk, 'name': item3.name, 'enabled': item3.enabled})
        self.assertEquals(len(changed), 0)
        self.assertEquals(len(removed), 1)
        self.assertEquals(removed[0]['data'], expected_serialized_item)

    def test_conditions(self):
        observer = QueryObserver(self.request(views.ExampleItemViewSet, enabled=True))
        items = observer.evaluate()

        self.assertEquals(observer.id, '0c2544b340aeb1919180ee6898a8e117de76b4a09dcedff7d6d172f3caa677c2')
        self.assertEquals(len(items), 0)

        item = models.ExampleItem()
        item.name = 'Example'
        item.enabled = False
        item.save()

        shortcuts.assign_perm('rest_framework_reactive.view_exampleitem', auth_models.AnonymousUser(), item)

        added, changed, removed = observer.evaluate(return_emitted=True)

        self.assertEquals(len(added), 0)
        self.assertEquals(len(changed), 0)
        self.assertEquals(len(removed), 0)

        item.enabled = True
        item.save()

        added, changed, removed = observer.evaluate(return_emitted=True)

        self.assertEquals(len(added), 1)
        self.assertEquals(added[0]['data'], {'id': item.pk, 'name': item.name, 'enabled': item.enabled})
        self.assertEquals(len(changed), 0)
        self.assertEquals(len(removed), 0)

    def test_joins(self):
        # Create some items so that we get a valid query (otherwise the query would be empty as django-guardian
        # would discover that the user doesn't have permissions to get any items).
        item = models.ExampleItem()
        item.name = 'Example'
        item.enabled = False
        item.save()

        subitem = models.ExampleSubItem(parent=item, enabled=True)
        subitem.save()

        shortcuts.assign_perm('rest_framework_reactive.view_exampleitem', auth_models.AnonymousUser(), item)
        shortcuts.assign_perm('rest_framework_reactive.view_examplesubitem', auth_models.AnonymousUser(), subitem)

        observer = QueryObserver(self.request(views.ExampleSubItemViewSet, parent__enabled=True))
        items = observer.evaluate()

        self.assertEquals(observer.id, '009955bdb64c21f679a7dfa2f747d6025ffef3185e5e01805662ebe8ca89c1d3')
        self.assertEquals(len(items), 0)

        observer_state = observer_models.Observer.objects.get(pk=observer.id)
        dependencies = observer_state.dependencies.all().values_list('table', flat=True)
        self.assertIn('rest_framework_reactive_tests_exampleitem', dependencies)
        self.assertIn('rest_framework_reactive_tests_examplesubitem', dependencies)

    def test_aggregations(self):
        item = models.ExampleItem()
        item.name = 'Example'
        item.enabled = False
        item.save()

        m2m_item = models.ExampleM2MItem()
        m2m_item.save()

        shortcuts.assign_perm('rest_framework_reactive.view_exampleitem', auth_models.AnonymousUser(), item)

        observer = QueryObserver(self.request(views.AggregationTestViewSet, items=[m2m_item.pk]))
        observer.evaluate()

        # There should be a dependency on the intermediate table.
        observer_state = observer_models.Observer.objects.get(pk=observer.id)
        dependencies = observer_state.dependencies.all().values_list('table', flat=True)
        self.assertIn('rest_framework_reactive_tests_examplem2mitem_items', dependencies)

    def test_order(self):
        observer = QueryObserver(self.request(views.ExampleItemViewSet, ordering='name'))
        items = observer.evaluate()

        self.assertEquals(len(items), 0)

        item = models.ExampleItem()
        item.name = 'D'
        item.enabled = False
        item.save()

        shortcuts.assign_perm('rest_framework_reactive.view_exampleitem', auth_models.AnonymousUser(), item)

        added, changed, removed = observer.evaluate(return_emitted=True)

        self.assertEquals(len(added), 1)
        self.assertEquals(added[0]['data'], {'id': item.pk, 'name': item.name, 'enabled': item.enabled})
        self.assertEquals(added[0]['order'], 0)
        self.assertEquals(len(changed), 0)
        self.assertEquals(len(removed), 0)

        item2 = models.ExampleItem()
        item2.name = 'A'
        item2.enabled = True
        item2.save()

        shortcuts.assign_perm('rest_framework_reactive.view_exampleitem', auth_models.AnonymousUser(), item2)

        added, changed, removed = observer.evaluate(return_emitted=True)

        self.assertEquals(len(added), 1)
        self.assertEquals(added[0]['data'], {'id': item2.pk, 'name': item2.name, 'enabled': item2.enabled})
        self.assertEquals(added[0]['order'], 0)
        # Check that the first item has changed, because its order has changed.
        self.assertEquals(len(changed), 1)
        self.assertEquals(changed[0]['data'], {'id': item.pk, 'name': item.name, 'enabled': item.enabled})
        self.assertEquals(changed[0]['order'], 1)
        self.assertEquals(len(removed), 0)

        item3 = models.ExampleItem()
        item3.name = 'C'
        item3.enabled = True
        item3.save()

        shortcuts.assign_perm('rest_framework_reactive.view_exampleitem', auth_models.AnonymousUser(), item3)

        added, changed, removed = observer.evaluate(return_emitted=True)
        self.assertEquals(len(added), 1)
        self.assertEquals(added[0]['data'], {'id': item3.pk, 'name': item3.name, 'enabled': item3.enabled})
        self.assertEquals(added[0]['order'], 1)
        self.assertEquals(len(changed), 1)
        self.assertEquals(changed[0]['data'], {'id': item.pk, 'name': item.name, 'enabled': item.enabled})
        self.assertEquals(changed[0]['order'], 2)
        self.assertEquals(len(removed), 0)



# Channel used for notifying workers from ORM signals.
CHANNEL_WORKER_NOTIFY = 'rest_framework_reactive.worker'
# Channel used for scheduling polling of observers.
CHANNEL_POLL_OBSERVER = 'rest_framework_reactive.poll_observer'
# Group used for individual sessions.
GROUP_SESSIONS = 'rest_framework_reactive.session.{session_id}'

# Message type for ORM table change notifications.
TYPE_ORM_NOTIFY_TABLE = 'orm.notify_table'
# Message type for polling observable evaluation.
TYPE_POLL_OBSERVER = 'poll_observer'
# Message type for observer item updates.
TYPE_ITEM_UPDATE = 'observer.update'

ORM_NOTIFY_KIND_CREATE = 'create'
ORM_NOTIFY_KIND_UPDATE = 'update'
ORM_NOTIFY_KIND_DELETE = 'delete'



import inspect

from rest_framework import response, viewsets

from django.db import transaction

from . import observer, request as observer_request


def observable(method_or_viewset):
    """Make the specified ViewSet or ViewSet method  observable.

    Decorating a ViewSet class is the same as decorating its `list` method.

    If the decorated method returns a response containing a list of items, it
    must use the provided `LimitOffsetPagination` for any pagination. In case
    a non-list response is returned, the resulting item will be wrapped into
    a list.

    When multiple decorators are used, `observable` must be the first one to be
    applied as it needs access to the method name.
    """

    if inspect.isclass(method_or_viewset):
        list_method = getattr(method_or_viewset, 'list', None)
        if list_method is not None:
            method_or_viewset.list = observable(list_method)

        return method_or_viewset

    # Do not decorate an already observable method twice.
    if getattr(method_or_viewset, 'is_observable', False):
        return method_or_viewset

    def wrapper(self, request, *args, **kwargs):
        if observer_request.OBSERVABLE_QUERY_PARAMETER in request.query_params:
            # TODO: Validate the session identifier.
            session_id = request.query_params[observer_request.OBSERVABLE_QUERY_PARAMETER]

            # Create request and subscribe the session to given observer.
            request = observer_request.Request(self.__class__, method_or_viewset.__name__, request, args, kwargs)

            # Create and evaluate observer.
            instance = observer.QueryObserver(request)
            with transaction.atomic():
                data = instance.evaluate()
                observer.add_subscriber(session_id, instance.id)

            return response.Response({
                'observer': instance.id,
                'items': data,
            })
        else:
            # Non-reactive API.
            return method_or_viewset(self, request, *args, **kwargs)

    wrapper.is_observable = True

    # Copy over any special observable attributes.
    for attribute in dir(method_or_viewset):
        if attribute.startswith(observer.OBSERVABLE_OPTIONS_PREFIX):
            setattr(wrapper, attribute, getattr(method_or_viewset, attribute))

    return wrapper


def primary_key(name):
    """Set primary key for tracking observable items.

    :param name: Name of the primary key field
    """

    def decorator(method):
        method.observable_primary_key = name
        return method

    return decorator


def polling_observable(interval):
    """Set polling interval for a polling observable.

    A decorator which configures the given observable as a polling
    observable. Instead of tracking changes based on notifications from
    the ORM, the observer is polled periodically.

    :param interval: Poll interval
    """

    def decorator(method):
        method.observable_change_detection = observer.Options.CHANGE_DETECTION_POLL
        method.observable_poll_interval = interval
        return method

    return decorator



from django.conf import settings


def get_queryobserver_settings():
    """Query observer connection configuration."""
    defaults = {
        # Observers going over these limits will emit warnings.
        'warnings': {
            'max_result_length': 1000,
            'max_processing_time': 1.0,
        },
        # Observers going over these limits will be stopped.
        'errors': {
            'max_processing_time': 20.0,
        },
        # Update batch delay (in seconds). If a new update comes earlier than the
        # delay value, queue processing will be delayed so multiple updates can be
        # batched. A higher value introduces more latency.
        'update_batch_delay': 5,
    }
    defaults.update(getattr(settings, 'DJANGO_REST_FRAMEWORK_REACTIVE', {}))
    return defaults



"""The views defined here are only used during testing."""
import time

from rest_framework import mixins, viewsets
from rest_framework.pagination import LimitOffsetPagination
from rest_framework.response import Response

from rest_framework_reactive.decorators import observable, polling_observable

from . import models, serializers


@observable
class ExampleItemViewSet(mixins.RetrieveModelMixin,
                         mixins.ListModelMixin,
                         viewsets.GenericViewSet):

    queryset = models.ExampleItem.objects.all()
    serializer_class = serializers.ExampleItemSerializer
    filter_fields = ('name', 'enabled', 'items')


@observable
class ExampleSubItemViewSet(mixins.RetrieveModelMixin,
                            mixins.ListModelMixin,
                            viewsets.GenericViewSet):

    queryset = models.ExampleSubItem.objects.all()
    serializer_class = serializers.ExampleSubItemSerializer
    filter_fields = ('parent__enabled', 'enabled')


@observable
class PaginatedViewSet(mixins.RetrieveModelMixin,
                       mixins.ListModelMixin,
                       viewsets.GenericViewSet):

    queryset = models.ExampleItem.objects.all()
    serializer_class = serializers.ExampleItemSerializer
    pagination_class = LimitOffsetPagination


@observable
class AggregationTestViewSet(mixins.RetrieveModelMixin,
                             mixins.ListModelMixin,
                             viewsets.GenericViewSet):

    queryset = models.ExampleItem.objects.all()
    serializer_class = serializers.ExampleItemSerializer
    filter_fields = ('items',)

    def list(self, request, *args, **kwargs):
        """Filtered query, which just returns a count.

        Such a formulation is used to force the compiler to generate a
        subquery, which uses the M2M relation.
        """
        queryset = self.filter_queryset(self.get_queryset())
        return Response({
            'count': queryset.count(),
        })


class PollingObservableViewSet(mixins.RetrieveModelMixin,
                               mixins.ListModelMixin,
                               viewsets.GenericViewSet):

    queryset = models.ExampleItem.objects.none()
    serializer_class = serializers.ExampleItemSerializer

    @observable
    @polling_observable(5)
    def list(self, request, *args, **kwargs):
        return Response({'static': 'This is a polling observable: {}'.format(time.time())})



import pytest

from django.conf.urls import url

from channels.db import database_sync_to_async
from channels.testing import ApplicationCommunicator, WebsocketCommunicator
from channels.routing import URLRouter
from channels.layers import get_channel_layer
from rest_framework import test as api_test, request as api_request

from rest_framework_reactive import request as observer_request, models as observer_models
from rest_framework_reactive.consumers import ClientConsumer, PollObserversConsumer, WorkerConsumer
from rest_framework_reactive.protocol import *
from rest_framework_reactive.observer import QueryObserver, add_subscriber, remove_subscriber

from . import models, views

# Create test request factory.
factory = api_test.APIRequestFactory()


def create_request(viewset_class, **kwargs):
    request = observer_request.Request(
        viewset_class,
        'list',
        api_request.Request(factory.get('/', kwargs))
    )

    return request


@pytest.mark.django_db(transaction=True)
@pytest.mark.asyncio
async def test_worker_and_client():
    client_consumer = URLRouter([
        url(r'^ws/(?P<subscriber_id>.+)$', ClientConsumer),
    ])

    client = WebsocketCommunicator(client_consumer, '/ws/test-session')
    worker = ApplicationCommunicator(WorkerConsumer, {
        'type': 'channel',
        'channel': CHANNEL_WORKER_NOTIFY,
    })

    # Connect client.
    status, _ = await client.connect()
    assert status is True

    # Create an observer.
    @database_sync_to_async
    def create_observer():
        observer = QueryObserver(create_request(views.PaginatedViewSet, offset=0, limit=10))
        items = observer.evaluate()
        assert not items

        add_subscriber('test-session', observer.id)

    await create_observer()

    # Create a single model instance for the observer model.
    @database_sync_to_async
    def create_model():
        return models.ExampleItem.objects.create(enabled=True, name="hello world").pk

    primary_key = await create_model()

    await worker.send_input({
        'type': TYPE_ORM_NOTIFY_TABLE,
        'table': models.ExampleItem._meta.db_table,
        'kind': ORM_NOTIFY_KIND_CREATE,
        'primary_key': primary_key,
    })

    response = await client.receive_json_from()
    assert response['msg'] == 'added'
    assert response['primary_key'] == 'id'
    assert response['order'] == 0
    assert response['item'] == {'id': 1, 'enabled': True, 'name': 'hello world'}

    # No other messages should be sent.
    assert await client.receive_nothing() is True

    await client.disconnect()

    # Ensure that subscriber and observer have been removed.
    @database_sync_to_async
    def check_subscribers():
        assert observer_models.Subscriber.objects.all().count() == 0
        assert observer_models.Observer.objects.all().count() == 0

    await check_subscribers()


@pytest.mark.asyncio
async def test_poll_observer():
    poller = ApplicationCommunicator(PollObserversConsumer, {
        'type': 'channel',
        'channel': CHANNEL_POLL_OBSERVER,
    })

    await poller.send_input({
        'type': TYPE_POLL_OBSERVER,
        'observer': 'test',
        'interval': 5,
    })

    # Nothing should be received in the frist 4 seconds.
    await poller.receive_nothing(timeout=4)
    # Then after two more seconds we should get a notification.
    notify = await poller.receive_output(timeout=2)
    assert notify['type'] == TYPE_POLL_OBSERVER
    assert notify['observer'] == 'test'


@pytest.mark.django_db(transaction=True)
@pytest.mark.asyncio
async def test_poll_observer_integration():
    client_consumer = URLRouter([
        url(r'^ws/(?P<subscriber_id>.+)$', ClientConsumer),
    ])

    client = WebsocketCommunicator(client_consumer, '/ws/test-session')
    worker = ApplicationCommunicator(WorkerConsumer, {
        'type': 'channel',
        'channel': CHANNEL_WORKER_NOTIFY,
    })
    poller = ApplicationCommunicator(PollObserversConsumer, {
        'type': 'channel',
        'channel': CHANNEL_POLL_OBSERVER,
    })

    # Connect client.
    status, _ = await client.connect()
    assert status is True

    # Create an observer.
    @database_sync_to_async
    def create_observer():
        observer = QueryObserver(create_request(views.PollingObservableViewSet))
        items = observer.evaluate()
        assert len(items) == 1

        add_subscriber('test-session', observer.id)
        return observer.id

    observer_id = await create_observer()

    # Ensure that a notification message was sent to the poller.
    channel_layer = get_channel_layer()

    notify = await channel_layer.receive(CHANNEL_POLL_OBSERVER)
    assert notify['type'] == TYPE_POLL_OBSERVER
    assert notify['interval'] == 5
    assert notify['observer'] == observer_id

    # Dispatch notification to poller as our poller uses a dummy queue.
    await poller.send_input(notify)
    # Nothing should be received in the frist 4 seconds.
    await poller.receive_nothing(timeout=4)
    # Then after two more seconds we should get a notification.
    notify = await poller.receive_output(timeout=2)

    # Dispatch notification to worker.
    await worker.send_input(notify)

    # Ensure another notification message was sent to the poller.
    notify = await channel_layer.receive(CHANNEL_POLL_OBSERVER)
    assert notify['type'] == TYPE_POLL_OBSERVER
    assert notify['interval'] == 5
    assert notify['observer'] == observer_id

    # Ensure client got notified of changes.
    response = await client.receive_json_from()
    assert response['msg'] == 'changed'
    assert response['primary_key'] == 'id'
    assert response['order'] == 0
    assert response['item']['static'].startswith('This is a polling observable:')

    # No other messages should be sent.
    assert await client.receive_nothing() is True

    await client.disconnect()



import asyncio
import pickle

from channels.consumer import AsyncConsumer, SyncConsumer
from channels.generic.websocket import JsonWebsocketConsumer

from .models import Observer, Subscriber
from .observer import QueryObserver
from .protocol import GROUP_SESSIONS, TYPE_POLL_OBSERVER


class PollObserversConsumer(AsyncConsumer):
    """Consumer for polling observers."""

    async def poll_observer(self, message):
        """Poll observer after a delay."""
        # Sleep until we need to notify the observer.
        await asyncio.sleep(message['interval'])

        # Dispatch task to evaluate the observable.
        await self.send({
            'type': TYPE_POLL_OBSERVER,
            'observer': message['observer'],
        })


class WorkerConsumer(SyncConsumer):
    """Worker consumer."""

    def __init__(self, *args, **kwargs):
        """Construct observer worker consumer."""
        self._executor_cache = {}
        super().__init__(*args, **kwargs)

    def _get_executor(self, state):
        """Get executor for given observer's state."""
        try:
            return self._executor_cache[state.pk]
        except KeyError:
            executor = QueryObserver(pickle.loads(state.request))
            self._executor_cache[state.pk] = executor
            return executor

    def _evaluate(self, state):
        """Evaluate observer based on state."""
        # Load state into an executor.
        executor = self._get_executor(state)

        # Evaluate observer.
        executor.evaluate(return_full=False)

    def orm_notify_table(self, message):
        """Process notification from ORM."""
        # Find all observers with dependencies on the given table. Instantiate it immediately
        # to prevent us holding any locks on the table.
        observers = list(
            Observer.objects.filter(dependencies__table=message['table'])
                .distinct()
                .only('pk', 'request')
        )

        for state in observers:
            self._evaluate(state)

    def poll_observer(self, message):
        """Evaluate poll observer."""
        try:
            observer = Observer.objects.get(pk=message['observer'])
        except Observer.DoesNotExist:
            return

        self._evaluate(observer)


class ClientConsumer(JsonWebsocketConsumer):
    """Client consumer."""

    def websocket_connect(self, message):
        """Called when WebSocket connection is established."""
        self.session_id = self.scope['url_route']['kwargs']['subscriber_id']
        super().websocket_connect(message)

        # Create new subscriber object.
        Subscriber.objects.create(session_id=self.session_id)

    @property
    def groups(self):
        """Groups this channel should add itself to."""
        if not hasattr(self, 'session_id'):
            return []

        return [GROUP_SESSIONS.format(session_id=self.session_id)]

    def disconnect(self, code):
        """Called when WebSocket connection is closed."""
        Subscriber.objects.filter(session_id=self.session_id).delete()

    def observer_update(self, message):
        """Called when update from observer is received."""
        message.pop('type')

        # Demultiplex observer update into multiple messages.
        for action in ('added', 'changed', 'removed'):
            for item in message[action]:
                self.send_json({
                    'msg': action,
                    'observer': message['observer'],
                    'primary_key': message['primary_key'],
                    'order': item['order'],
                    'item': item['data'],
                })



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-05-12 12:00
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Dependency',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('table', models.CharField(max_length=100)),
            ],
        ),
        migrations.CreateModel(
            name='Item',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('primary_key', models.CharField(max_length=200)),
                ('order', models.IntegerField()),
                ('data', django.contrib.postgres.fields.jsonb.JSONField()),
            ],
            options={
                'ordering': ('observer', 'order'),
            },
        ),
        migrations.CreateModel(
            name='Observer',
            fields=[
                ('id', models.CharField(max_length=64, primary_key=True, serialize=False)),
                ('status', models.CharField(choices=[('new', 'New'), ('observing', 'Observing'), ('stopped', 'Stopped')], default='new', max_length=15)),
                ('request', models.BinaryField()),
                ('last_evaluation', models.DateTimeField(auto_now_add=True)),
                ('poll_interval', models.IntegerField(null=True)),
            ],
        ),
        migrations.CreateModel(
            name='Subscriber',
            fields=[
                ('session_id', models.CharField(max_length=100, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.AddField(
            model_name='observer',
            name='subscribers',
            field=models.ManyToManyField(to='rest_framework_reactive.Subscriber'),
        ),
        migrations.AddField(
            model_name='item',
            name='observer',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='items', to='rest_framework_reactive.Observer'),
        ),
        migrations.AddField(
            model_name='dependency',
            name='observer',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='dependencies', to='rest_framework_reactive.Observer'),
        ),
        migrations.AlterUniqueTogether(
            name='item',
            unique_together=set([('observer', 'primary_key'), ('observer', 'order')]),
        ),
        migrations.AlterUniqueTogether(
            name='dependency',
            unique_together=set([('observer', 'table')]),
        ),
    ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Making Django REST Framework reactive.

See:
https://github.com/genialis/django-rest-framework-reactive
"""

from setuptools import find_packages, setup
# Use codecs' open for a consistent encoding
from codecs import open
from os import path

base_dir = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(base_dir, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

# Get package metadata from 'rest_framework_reactive.__about__.py' file
about = {}
with open(path.join(base_dir, 'rest_framework_reactive', '__about__.py'), encoding='utf-8') as f:
    exec(f.read(), about)

setup(
    name=about['__title__'],

    version=about['__version__'],

    description=about['__summary__'],
    long_description=long_description,

    url=about['__url__'],

    author=about['__author__'],
    author_email=about['__email__'],

    license=about['__license__'],

    packages=find_packages(exclude=['tests']),
    python_requires='>=3.6',
    install_requires=[
        'Django~=1.11.6',
        'djangorestframework>=3.4.0',
        'channels~=2.1.1',
    ],
    extras_require={
        'docs': [
            'sphinx>=1.3.2',
            'sphinx_rtd_theme',
        ],
        'package': [
            'twine',
            'wheel',
        ],
        'test': [
            'djangorestframework-filters>=0.9.1',
            'django-guardian>=1.4.2',
            'django-jenkins>=0.17.0',
            'channels-redis~=2.1.0',
            'coverage>=3.7.1',
            'pep8>=1.6.2',
            'pylint>=1.4.3',
            'pytest~=3.5.1',
            'pytest-django~=3.2.1',
            'pytest-asyncio~=0.8.0',
            'psycopg2>=2.5.0',
            'check-manifest',
            'readme',
        ],
    },

    classifiers=[
        'Development Status :: 4 - Beta',

        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',

        'License :: OSI Approved :: Apache Software License',

        'Operating System :: OS Independent',

        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],

    keywords='django-rest-framework reactive django',
)



from __future__ import absolute_import, division, print_function, unicode_literals

from django import apps


class QueryObserverTestsConfig(apps.AppConfig):
    name = 'rest_framework_reactive.tests'
    label = 'rest_framework_reactive_tests'



import contextlib
import threading

from django.db.models.sql import compiler


class QueryInterceptor(object):
    """Django ORM query interceptor."""

    def __init__(self):
        self.intercepting_queries = 0
        self.tables = set()

    def _patch(self):
        """Monkey patch the SQLCompiler class to get all the referenced tables in a code block."""
        self.intercepting_queries += 1
        if self.intercepting_queries > 1:
            return

        self._original_as_sql = compiler.SQLCompiler.as_sql

        def as_sql(compiler, *args, **kwargs):
            try:
                return self._original_as_sql(compiler, *args, **kwargs)
            finally:
                self.tables.update(compiler.query.tables)

        compiler.SQLCompiler.as_sql = as_sql

    def _unpatch(self):
        """Restore SQLCompiler monkey patches."""

        self.intercepting_queries -= 1
        assert self.intercepting_queries >= 0

        if self.intercepting_queries:
            return

        compiler.SQLCompiler.as_sql = self._original_as_sql

    @contextlib.contextmanager
    def intercept(self, tables):
        """Intercept all tables used inside a codeblock.

        :param tables: Output tables set
        """
        self._patch()

        try:
            # Run the code block.
            yield
        finally:
            self._unpatch()
            # Collect intercepted tables.
            tables.update(self.tables)



from django.conf.urls import url

from channels.routing import ChannelNameRouter, ProtocolTypeRouter, URLRouter

from .consumers import ClientConsumer, PollObserversConsumer, WorkerConsumer
from .protocol import CHANNEL_POLL_OBSERVER, CHANNEL_WORKER_NOTIFY

application = ProtocolTypeRouter({
    # Client-facing consumers.
    'websocket': URLRouter([
        # To change the prefix, you can import ClientConsumer in your custom
        # Channels routing definitions instead of using these defaults.
        url(r'^ws/(?P<subscriber_id>.+)$', ClientConsumer),
    ]),

    # Background worker consumers.
    'channel': ChannelNameRouter({
        CHANNEL_POLL_OBSERVER: PollObserversConsumer,
        CHANNEL_WORKER_NOTIFY: WorkerConsumer,
    })
})
